@@ -412,7 +412,8 @@ httpClientHandler(int status,
         int rc = 0;
         if(!(connection->flags & CONN_BIGREQBUF))
             rc = httpConnectionBigifyReqbuf(connection);
-        if(rc > 0) {
+        if((connection->flags & CONN_BIGREQBUF) &&
+           connection->reqlen < bigBufferSize) {
             do_stream(IO_READ, connection->fd, connection->reqlen,
                       connection->reqbuf, bigBufferSize,
                       httpClientHandler, connection);
@@ -771,7 +772,6 @@ httpClientRequest(HTTPRequestPtr request, AtomPtr url)
             expect = NULL;
         }
         if(code) {
-            if(expect) releaseAtom(expect);
             request->flags |= REQUEST_FORCE_ERROR;
             httpClientDiscardBody(connection);
             httpClientNoticeErrorHeaders(request, code, message, challenge);
@@ -785,8 +785,14 @@ httpClientRequest(HTTPRequestPtr request, AtomPtr url)
     }
 
     if(expect) {
-        if(expect == atom100Continue) {
+        if(expect == atom100Continue && REQUEST_SIDE(request)) {
             request->flags |= REQUEST_WAIT_CONTINUE;
+        } else {
+            httpClientDiscardBody(connection);
+            httpClientNoticeError(request, 417,
+                                  internAtom("Expectation failed"));
+            releaseAtom(expect);
+            return 1;
         }
         releaseAtom(expect);
     }
@@ -817,6 +823,7 @@ httpClientRequest(HTTPRequestPtr request, AtomPtr url)
             httpClientNoticeError(request, 500,
                                   internAtom("CONNECT over big buffer "
                                              "not supported"));
+            return 1;
         }
         connection->flags &= ~CONN_READER;
         do_tunnel(connection->fd, connection->reqbuf, 
@@ -1244,6 +1251,9 @@ httpClientNoticeRequest(HTTPRequestPtr request, int novalidate)
     if(!mindlesslyCacheVary && (request->object->cache_control & CACHE_VARY))
         conditional = conditional && (request->object->etag != NULL);
 
+    conditional =
+        conditional && !(request->object->cache_control & CACHE_MISMATCH);
+
     request->object->flags |= OBJECT_VALIDATING;
     rc = request->object->request(request->object,
                                   conditional ? METHOD_CONDITIONAL_GET : 
