@@ -1001,10 +1001,12 @@ httpServerDoSide(HTTPConnectionPtr connection)
         connection->reqbuf = NULL;
         connection->reqlen = 0;
         pokeFdEvent(connection->fd, -ESHUTDOWN, POLLIN);
-        client->flags |= CONN_SIDE_READER;
-        do_stream(IO_READ | IO_IMMEDIATE,
-                  client->fd, 0, NULL, 0,
-                  httpClientSideHandler, client);
+        if(client->flags & CONN_READER) {
+            client->flags |= CONN_SIDE_READER;
+            do_stream(IO_READ | IO_IMMEDIATE | IO_NOTNOW,
+                      client->fd, 0, NULL, 0,
+                      httpClientSideHandler, client);
+        }
     } else if(!(request->flags & REQUEST_WAIT_CONTINUE) && doflush) {
         /* Make sure there's a reqbuf, as httpServerFinish uses
            it to determine if there's a writer. */
@@ -1146,8 +1148,7 @@ httpServerContinueConditionHandler(int status, ConditionHandlerPtr chandler)
     return 1;
 }
 
-/* s is 0 to keep the connection alive, 1 to shutdown the connection,
-   and -1 to keep the connection alive and keep the current request. */
+/* s is 0 to keep the connection alive, 1 to shutdown the connection */
 void
 httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
 {
@@ -1162,7 +1163,7 @@ httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
         assert(connection->pipelined == 0);
     }
 
-    if(s == 0 && (!connection->request ||
+    if(!s && (!connection->request ||
                   !(connection->request->flags & REQUEST_PERSISTENT)))
         s = 1;
 
@@ -1182,7 +1183,7 @@ httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
         goto done;
     }
 
-    if(s >= 0 && request) {
+    if(request) {
         /* Update statistics about the server */
         int size = -1, d = -1, rtt = -1, rate = -1;
         if(connection->offset > 0 && request->from >= 0)
@@ -1228,7 +1229,7 @@ httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
 
     assert(offset <= connection->len);
 
-    if(s <= 0) {
+    if(!s) {
         if(offset < connection->len) {
             assert(connection->buf != NULL);
             if(!connection->pipelined) {
@@ -1253,7 +1254,7 @@ httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
     connection->server->time = current_time.tv_sec;
     connection->serviced++;
 
-    if(s > 0) {
+    if(s) {
         if(connection->timeout)
             cancelTimeEvent(connection->timeout);
         connection->timeout = NULL;
@@ -1311,7 +1312,7 @@ httpServerFinish(HTTPConnectionPtr connection, int s, int offset)
            (server->pipeline == 3)) {
             server->pipeline++;
         }
-        if(s < 0 || connection->pipelined) {
+        if(connection->pipelined) {
             httpServerReply(connection, 1);
         } else {
             httpConnectionDestroyBuf(connection);
@@ -1930,13 +1931,19 @@ httpServerHandlerHeaders(int eof,
         return 1;
     }
 
+    if(date < 0)
+        date = current_time.tv_sec;
+
     if(code == 100) {
         releaseAtom(url);
         releaseAtom(message);
         /* We've already reset wait_continue above, but we must still
            ensure that the writer notices. */
         notifyObject(request->object);
-        httpServerFinish(connection, -1, rc);
+        connection->len -= rc;
+        if(connection->len > 0)
+            memmove(connection->buf, connection->buf + rc, connection->len);
+        httpServerReply(connection, 1);
         return 1;
     }
 
